@@ -4,15 +4,19 @@
 from collections import defaultdict
 from decimal import Decimal
 
-from trytond.i18n import gettext
-from trytond.model import ModelSQL, ModelView, Workflow, fields
-from trytond.model import DeactivableMixin
-from trytond.model.exceptions import AccessError
-from trytond.pyson import Eval, If, Bool
-from trytond.pool import Pool
-from trytond.transaction import Transaction
+from sql.functions import CharLength
 
+from trytond.i18n import gettext
+from trytond.model import (
+    DeactivableMixin, Index, ModelSQL, ModelView, Workflow, fields)
+from trytond.model.exceptions import AccessError
+from trytond.modules.company.model import (
+    employee_field, reset_employee, set_employee)
+from trytond.modules.currency.fields import Monetary
 from trytond.modules.product import price_digits
+from trytond.pool import Pool
+from trytond.pyson import Bool, Eval, If
+from trytond.transaction import Transaction
 
 
 class Type(DeactivableMixin, ModelSQL, ModelView):
@@ -33,47 +37,87 @@ class Complaint(Workflow, ModelSQL, ModelView):
     _states = {
         'readonly': Eval('state') != 'draft',
         }
-    _depends = ['state']
 
-    number = fields.Char('Number', readonly=True, select=True)
-    reference = fields.Char('Reference', select=True)
-    date = fields.Date('Date', states=_states, depends=_depends)
+    number = fields.Char("Number", readonly=True)
+    reference = fields.Char("Reference")
+    date = fields.Date('Date', states=_states)
     customer = fields.Many2One(
         'party.party', "Customer", required=True, states=_states,
         context={
             'company': Eval('company', -1),
             },
-        depends=_depends + ['company'])
-    address = fields.Many2One('party.address', 'Address',
-        domain=[('party', '=', Eval('customer'))],
-        states=_states, depends=_depends + ['customer'])
-    company = fields.Many2One('company.company', 'Company', required=True,
+        depends={'company'})
+    company = fields.Many2One(
+        'company.company', 'Company', required=True,
         states={
             'readonly': _states['readonly'] | Eval('origin'),
-            },
-        depends=_depends + ['origin'])
-    employee = fields.Many2One('company.employee', 'Employee',
-        states=_states, depends=_depends)
+            })
     type = fields.Many2One('sale.complaint.type', 'Type', required=True,
-        states=_states, depends=_depends)
+        states=_states)
     origin = fields.Reference('Origin', selection='get_origin',
+        domain={
+            'sale.sale': [
+                If(Eval('customer'),
+                    ('party', '=', Eval('customer')),
+                    ()),
+                ('company', '=', Eval('company')),
+                ('state', 'in', ['confirmed', 'processing', 'done']),
+                ],
+            'sale.line': [
+                ('type', '=', 'line'),
+                If(Eval('customer'),
+                    ('sale.party', '=', Eval('customer')),
+                    ()),
+                ('sale.company', '=', Eval('company')),
+                ('sale.state', 'in', ['confirmed', 'processing', 'done']),
+                ],
+            'account.invoice': [
+                If(Eval('customer'),
+                    ('party', '=', Eval('customer')),
+                    ()),
+                ('company', '=', Eval('company')),
+                ('type', '=', 'out'),
+                ('state', 'in', ['posted', 'paid']),
+                ],
+            'account.invoice.line': [
+                ('type', '=', 'line'),
+                If(Eval('customer'),
+                    ('invoice.party', '=', Eval('customer')),
+                    ()),
+                ('invoice.company', '=', Eval('company')),
+                ('invoice.type', '=', 'out'),
+                ('invoice.state', 'in', ['posted', 'paid']),
+                ],
+            },
         states={
             'readonly': ((Eval('state') != 'draft')
                 | Bool(Eval('actions', [0]))),
             'required': Bool(Eval('origin_model')),
             },
-        depends=['state', 'customer', 'origin_model', 'company'])
+        depends={'origin_model'})
     origin_id = fields.Function(fields.Integer('Origin ID'),
         'on_change_with_origin_id')
     origin_model = fields.Function(fields.Char('Origin Model'),
         'on_change_with_origin_model')
-    description = fields.Text('Description', states=_states, depends=_depends)
+    description = fields.Text('Description', states=_states)
     actions = fields.One2Many('sale.complaint.action', 'complaint', 'Actions',
         states={
             'readonly': ((Eval('state') != 'draft')
                 | (If(~Eval('origin_id', 0), 0, Eval('origin_id', 0)) <= 0)),
             },
-        depends=['state', 'origin_model', 'origin_id'])
+        depends={'origin_model'})
+    submitted_by = employee_field(
+        "Submitted By",
+        states=['waiting', 'approved', 'rejected', 'done', 'cancelled'])
+    approved_by = employee_field(
+        "Approved By",
+        states=['approved', 'rejected', 'done', 'cancelled'])
+    rejected_by = employee_field(
+        "Rejected By",
+        states=['approved', 'rejected', 'done', 'cancelled'])
+    cancelled_by = employee_field(
+        "Cancelled By",
+        states=['cancelled'])
     state = fields.Selection([
             ('draft', 'Draft'),
             ('waiting', 'Waiting'),
@@ -81,11 +125,21 @@ class Complaint(Workflow, ModelSQL, ModelView):
             ('rejected', 'Rejected'),
             ('done', 'Done'),
             ('cancelled', 'Cancelled'),
-            ], 'State', readonly=True, required=True)
+            ], "State", readonly=True, required=True, sort=False)
 
     @classmethod
     def __setup__(cls):
+        cls.number.search_unaccented = False
+        cls.reference.search_unaccented = False
         super(Complaint, cls).__setup__()
+        t = cls.__table__()
+        cls._sql_indexes.update({
+                Index(t, (t.reference, Index.Similarity())),
+                Index(
+                    t,
+                    (t.state, Index.Equality()),
+                    where=t.state.in_(['draft', 'waiting', 'approved'])),
+                })
         cls._order.insert(0, ('date', 'DESC'))
         cls._transitions |= set((
                 ('draft', 'waiting'),
@@ -130,12 +184,6 @@ class Complaint(Workflow, ModelSQL, ModelView):
                     },
                 })
 
-        origin_domain = []
-        for model, domain in cls._origin_domains().items():
-            origin_domain = If(Eval('origin_model') == model,
-                domain, origin_domain)
-        cls.origin.domain = [origin_domain]
-
         actions_domains = cls._actions_domains()
         actions_domain = [('action', 'in', actions_domains.pop(None))]
         for model, actions in actions_domains.items():
@@ -147,50 +195,13 @@ class Complaint(Workflow, ModelSQL, ModelView):
     def __register__(cls, module_name):
         table_h = cls.__table_handler__(module_name)
 
-        # Migration from 3.8: rename reference into number
-        if (table_h.column_exist('reference')
-                and not table_h.column_exist('number')):
-            table_h.column_rename('reference', 'number')
+        # Migration from 6.4: rename employee into submitted_by
+        if (table_h.column_exist('employee')
+                and not table_h.column_exist('submitted_by')):
+            table_h.column_rename('employee', 'submitted_by')
 
         super(Complaint, cls).__register__(module_name)
 
-    @classmethod
-    def _origin_domains(cls):
-        return {
-            'sale.sale': [
-                If(Eval('customer'),
-                    ('party', '=', Eval('customer')),
-                    ()),
-                ('company', '=', Eval('company')),
-                ('state', 'in', ['confirmed', 'processing', 'done']),
-                ],
-            'sale.line': [
-                ('type', '=', 'line'),
-                If(Eval('customer'),
-                    ('sale.party', '=', Eval('customer')),
-                    ()),
-                ('sale.company', '=', Eval('company')),
-                ('sale.state', 'in', ['confirmed', 'processing', 'done']),
-                ],
-            'account.invoice': [
-                If(Eval('customer'),
-                    ('party', '=', Eval('customer')),
-                    ()),
-                ('company', '=', Eval('company')),
-                ('type', '=', 'out'),
-                ('state', 'in', ['posted', 'paid']),
-                ],
-            'account.invoice.line': [
-                ('type', '=', 'line'),
-                If(Eval('customer'),
-                    ('invoice.party', '=', Eval('customer')),
-                    ()),
-                ('invoice.company', '=', Eval('company')),
-                ('invoice.type', '=', 'out'),
-                ('invoice.state', 'in', ['posted', 'paid']),
-                ],
-            }
-
     @classmethod
     def _actions_domains(cls):
         return {
@@ -201,6 +212,11 @@ class Complaint(Workflow, ModelSQL, ModelView):
             'account.invoice.line': ['credit_note'],
             }
 
+    @classmethod
+    def order_number(cls, tables):
+        table, _ = tables[None]
+        return [CharLength(table.number), table.number]
+
     @staticmethod
     def default_date():
         pool = Pool()
@@ -278,6 +294,10 @@ class Complaint(Workflow, ModelSQL, ModelView):
         else:
             default = default.copy()
         default.setdefault('number', None)
+        default.setdefault('submitted_by')
+        default.setdefault('approved_by')
+        default.setdefault('rejected_by')
+        default.setdefault('cancelled_by')
         return super(Complaint, cls).copy(complaints, default=default)
 
     @classmethod
@@ -292,36 +312,44 @@ class Complaint(Workflow, ModelSQL, ModelView):
     @classmethod
     @ModelView.button
     @Workflow.transition('cancelled')
+    @set_employee('cancelled_by')
     def cancel(cls, complaints):
         pass
 
     @classmethod
     @ModelView.button
     @Workflow.transition('draft')
+    @reset_employee(
+        'submitted_by', 'approved_by', 'rejected_by', 'cancelled_by')
     def draft(cls, complaints):
         pass
 
     @classmethod
     @ModelView.button
     @Workflow.transition('waiting')
+    @set_employee('submitted_by')
     def wait(cls, complaints):
         pass
 
     @classmethod
     @ModelView.button
     @Workflow.transition('approved')
+    @set_employee('approved_by')
     def approve(cls, complaints):
         pool = Pool()
         Configuration = pool.get('sale.configuration')
+        transaction = Transaction()
+        context = transaction.context
         config = Configuration(1)
-        with Transaction().set_context(
-                queue_name='sale',
-                queue_scheduled_at=config.sale_process_after):
+        with transaction.set_context(
+                queue_scheduled_at=config.sale_process_after,
+                queue_batch=context.get('queue_batch', True)):
             cls.__queue__.process(complaints)
 
     @classmethod
     @ModelView.button
     @Workflow.transition('rejected')
+    @set_employee('rejected_by')
     def reject(cls, complaints):
         pass
 
@@ -355,18 +383,16 @@ class Action(ModelSQL, ModelView):
         'readonly': ((Eval('complaint_state') != 'draft')
             | Bool(Eval('result'))),
         }
-    _depends = ['complaint_state', 'result']
     _line_states = {
         'invisible': ~Eval('_parent_complaint', {}
             ).get('origin_model', 'sale.line').in_(
             ['sale.line', 'account.invoice.line']),
         'readonly': _states['readonly'],
         }
-    _line_depends = _depends
 
     complaint = fields.Many2One(
         'sale.complaint', 'Complaint', required=True, ondelete='CASCADE',
-        states=_states, depends=_depends)
+        states=_states)
     action = fields.Selection([
             ('sale_return', 'Create Sale Return'),
             ('credit_note', 'Create Credit Note'),
@@ -379,7 +405,6 @@ class Action(ModelSQL, ModelView):
                 ).get('origin_model', 'sale.sale') != 'sale.sale',
             'readonly': _states['readonly'],
             },
-        depends=_depends,
         help='Leave empty for all lines.')
 
     invoice_lines = fields.One2Many(
@@ -391,31 +416,26 @@ class Action(ModelSQL, ModelView):
                 ) != 'account.invoice',
             'readonly': _states['readonly'],
             },
-        depends=_depends,
         help='Leave empty for all lines.')
 
-    quantity = fields.Float('Quantity',
-        digits=(16, Eval('unit_digits', 2)),
-        states=_line_states, depends=_line_depends + ['unit_digits'],
+    quantity = fields.Float(
+        "Quantity", digits='unit',
+        states=_line_states,
         help='Leave empty for the same quantity.')
     unit = fields.Function(fields.Many2One('product.uom', 'Unit',
-            states=_line_states, depends=_line_depends),
+            states=_line_states),
         'on_change_with_unit')
-    unit_digits = fields.Function(
-        fields.Integer('Unit Digits'), 'on_change_with_unit_digits')
-    unit_price = fields.Numeric('Unit Price', digits=price_digits,
-        states=_line_states, depends=_line_depends,
+    unit_price = Monetary(
+        "Unit Price", currency='currency', digits=price_digits,
+        states=_line_states,
         help='Leave empty for the same price.')
 
-    amount = fields.Function(fields.Numeric(
-            "Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    amount = fields.Function(Monetary(
+            "Amount", 'currency', digits='currency'),
         'on_change_with_amount')
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"),
         'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer("Currency Digits"),
-        'on_change_with_currency_digits')
 
     result = fields.Reference('Result', selection='get_result', readonly=True)
 
@@ -434,15 +454,7 @@ class Action(ModelSQL, ModelView):
         if (self.complaint
                 and self.complaint.origin_model in {
                     'sale.line', 'account.invoice.line'}):
-            return self.complaint.origin.unit.id
-
-    @fields.depends('complaint',
-        '_parent_complaint.origin_model', '_parent_complaint.origin')
-    def on_change_with_unit_digits(self, name=None):
-        if (self.complaint
-                and self.complaint.origin_model in {
-                    'sale.line', 'account.invoice.line'}):
-            return self.complaint.origin.unit.digits
+            return self.complaint.origin.unit
 
     @fields.depends(
         'quantity', 'unit_price', 'currency', 'sale_lines', 'invoice_lines',
@@ -489,17 +501,7 @@ class Action(ModelSQL, ModelView):
                 and self.complaint.origin_model in {
                     'sale.sale', 'sale.line',
                     'account.invoice', 'account.invoice.line'}):
-            return self.complaint.origin.currency.id
-
-    @fields.depends(
-        'complaint',
-        '_parent_complaint.origin_model', '_parent_complaint.origin')
-    def on_change_with_currency_digits(self, name=None):
-        if (self.complaint
-                and self.complaint.origin_model in {
-                    'sale.sale', 'sale.line',
-                    'account.invoice', 'account.invoice.line'}):
-            return self.complaint.origin.currency.digits
+            return self.complaint.origin.currency
 
     @classmethod
     def get_complaint_states(cls):
@@ -607,6 +609,7 @@ class Action(ModelSQL, ModelView):
             with Transaction().set_context(_account_invoice_correction=True):
                 credit_note, = Invoice.copy([invoice], default={
                         'lines': [],
+                        'taxes': [],
                         })
                 # Copy each line one by one to get negative and positive lines
                 # following each other
@@ -641,38 +644,30 @@ class Action(ModelSQL, ModelView):
 
 
 class _Action_Line:
-
+    __slots__ = ()
     _states = {
         'readonly': (
             (Eval('complaint_state') != 'draft')
             | Bool(Eval('_parent_action.result', True))),
         }
-    _depends = ['complaint_state']
 
-    action = fields.Many2One('sale.complaint.action', 'Action',
-        ondelete='CASCADE', select=True, required=True)
+    action = fields.Many2One(
+        'sale.complaint.action', "Action", ondelete='CASCADE', required=True)
     quantity = fields.Float(
-        "Quantity",
-        digits=(16, Eval('unit_digits', 2)),
-        states=_states,
-        depends=_depends + ['unit_digits'])
+        "Quantity", digits='unit', states=_states)
     unit = fields.Function(
         fields.Many2One('product.uom', "Unit"), 'on_change_with_unit')
-    unit_digits = fields.Function(
-        fields.Integer("Unit Digits"), 'on_change_with_unit_digits')
-    unit_price = fields.Numeric(
-        "Unit Price", digits=price_digits, states=_states, depends=_depends,
+    unit_price = Monetary(
+        "Unit Price", currency='currency', digits=price_digits,
+        states=_states,
         help='Leave empty for the same price.')
 
-    amount = fields.Function(fields.Numeric(
-            "Amount", digits=(16, Eval('currency_digits', 2)),
-            depends=['currency_digits']),
+    amount = fields.Function(Monetary(
+            "Amount", currency='currency', digits='currency'),
         'on_change_with_amount')
     currency = fields.Function(fields.Many2One(
             'currency.currency', "Currency"),
         'on_change_with_currency')
-    currency_digits = fields.Function(fields.Integer("Currency Digits"),
-        'on_change_with_currency_digits')
 
     complaint_state = fields.Function(
         fields.Selection('get_complaint_states', "Complaint State"),
@@ -689,9 +684,6 @@ class _Action_Line:
     def on_change_with_unit(self, name=None):
         raise NotImplementedError
 
-    def on_change_with_unit_digits(self, name=None):
-        raise NotImplementedError
-
     @fields.depends('currency', methods=['get_quantity', 'get_unit_price'])
     def on_change_with_amount(self, name=None):
         quantity = self.get_quantity() or 0
@@ -709,13 +701,7 @@ class _Action_Line:
 
     @fields.depends('action', '_parent_action.currency')
     def on_change_with_currency(self, name=None):
-        if self.action and self.action.currency:
-            return self.action.currency.id
-
-    @fields.depends('action', '_parent_action.currency')
-    def on_change_with_currency_digits(self, name=None):
-        if self.action and self.action.currency:
-            return self.action.currency.digits
+        return self.action.currency if self.action else None
 
     @classmethod
     def get_complaint_states(cls):
@@ -746,18 +732,11 @@ class Action_SaleLine(_Action_Line, ModelView, ModelSQL):
         domain=[
             ('type', '=', 'line'),
             ('sale', '=', Eval('complaint_origin_id', -1)),
-            ],
-        depends=['complaint_origin_id'])
+            ])
 
     @fields.depends('line')
     def on_change_with_unit(self, name=None):
-        if self.line and self.line.unit:
-            return self.line.unit.id
-
-    @fields.depends('line')
-    def on_change_with_unit_digits(self, name=None):
-        if self.line:
-            return self.line.unit.digits
+        return self.line.unit if self.line else None
 
     @fields.depends('quantity', 'line')
     def get_quantity(self):
@@ -784,18 +763,11 @@ class Action_InvoiceLine(_Action_Line, ModelView, ModelSQL):
         domain=[
             ('type', '=', 'line'),
             ('invoice', '=', Eval('complaint_origin_id', -1)),
-            ],
-        depends=['complaint_origin_id'])
+            ])
 
     @fields.depends('line')
     def on_change_with_unit(self, name=None):
-        if self.line and self.line.unit:
-            return self.line.unit.id
-
-    @fields.depends('line')
-    def on_change_with_unit_digits(self, name=None):
-        if self.line:
-            return self.line.unit.digits
+        return self.line.unit if self.line else None
 
     @fields.depends('quantity', 'line')
     def get_quantity(self):
