@@ -1,9 +1,8 @@
 # This file is part of Tryton.  The COPYRIGHT file at the top level of
 # this repository contains the full copyright notices and license terms.
 
-from trytond import backend
-from trytond.pool import PoolMeta, Pool
 from trytond.model import fields
+from trytond.pool import Pool, PoolMeta
 from trytond.pyson import Eval, Id
 
 
@@ -11,7 +10,7 @@ class Configuration(metaclass=PoolMeta):
     __name__ = 'sale.configuration'
 
     complaint_sequence = fields.MultiValue(fields.Many2One(
-            'ir.sequence', "Complaint Sequence",
+            'ir.sequence', "Complaint Sequence", required=True,
             domain=[
                 ('sequence_type', '=',
                     Id('sale_complaint', 'sequence_type_complaint')),
@@ -35,32 +34,12 @@ class Configuration(metaclass=PoolMeta):
 class ConfigurationSequence(metaclass=PoolMeta):
     __name__ = 'sale.configuration.sequence'
     complaint_sequence = fields.Many2One(
-        'ir.sequence', "Complaint Sequence",
+        'ir.sequence', "Complaint Sequence", required=True,
         domain=[
             ('sequence_type', '=',
                 Id('sale_complaint', 'sequence_type_complaint')),
             ('company', 'in', [Eval('company', -1), None]),
-            ],
-        depends=['company'])
-
-    @classmethod
-    def __register__(cls, module_name):
-        exist = backend.TableHandler.table_exist(cls._table)
-        if exist:
-            table = cls.__table_handler__(module_name)
-            exist &= table.column_exist('complaint_sequence')
-
-        super(ConfigurationSequence, cls).__register__(module_name)
-
-        if not exist:
-            cls._migrate_property([], [], [])
-
-    @classmethod
-    def _migrate_property(cls, field_names, value_names, fields):
-        field_names.append('complaint_sequence')
-        value_names.append('complaint_sequence')
-        super(ConfigurationSequence, cls)._migrate_property(
-            field_names, value_names, fields)
+            ])
 
     @classmethod
     def default_complaint_sequence(cls):
